package PureMessage::QDigest::Cache;

use strict;
use Data::Dump qw(dump);
use Date::Parse qw(str2time);
use File::Path qw(mkpath);
use File::Basename qw(dirname);
use File::Find qw(find);

sub DELIM() { "\0" }
sub DRYDIR() { "dryrun" }

sub new {
    my ($class, $dir) = @_;
    # Create the directory if it doesn't exist, and croak if we can't.
    -d $dir or mkdir($dir, 0777) or die "can't mkdir '$dir': $!";
    bless {
	dir     => [$dir],
	dryrun  => 0,
	lock	=> "$dir/.lock",
    }, ref($class) || $class;
}

# Called if we're doing a dry run of publishing the digests. Mirrors all the
# caches into the dry-run area.
sub dry_run {
    my $self = shift;
    my $root = join "/", @{$self->{dir}};
    find({
	    wanted => sub {
		return unless -f _;
		my $src = $File::Find::name;
		(my $dst = $src) =~ s#^\Q$root\E#"$root/" . DRYDIR#e;
		my $dir = dirname($dst);
		-d $dir or mkpath($dir) or die "can't create dir $dir: $!";

		# Copy the cache files
		open(my $DST, '>', $dst) or die "can't write $dst: $!";
		open(my $SRC, '<', $src) or die "can't read $src: $!";
		binmode($DST);
		binmode($SRC);
		print $DST $_ while <$SRC>;
		close $DST or die "can't write $dst: $!";
		close $SRC;
	    },
	    preprocess => sub {
		return () if $_ eq DRYDIR;
		@_;
	    },
    }, $root);

    # Update the working directory to be the dryrun subdir.
    push @{$self->{dir}}, DRYDIR;
}

# file name encoding/decoding (based on similar code in List::Utils)
# include some checking to get a "safe" filename
sub addr2fname {
    my $addr = shift;
    $addr =~ s/\A\s*//; # trim space chars
    $addr =~ s/\s*\Z//; # trim space chars
    $addr =~ s/([^\w\@\-\.])/sprintf("=%02X", ord($1))/ge;
    return $addr;
}

sub fname2addr {
    my $fname = shift;
    $fname =~ s/=([\da-fA-F]{2})/pack("C", hex($1))/ge;
    return $fname;
}

sub get_path {
    my ($self, $recip) = @_;
    $recip = addr2fname($recip);
    my $xx = substr($recip, 0, 2);
    $xx =~ s/[^a-z\d]/x/g;
    return (@{$self->{dir}}, $xx, $recip);
}

sub recip2name {
    my ($self, $recip) = @_;
    return join('/', $self->get_path($recip)); 
}

sub load {
    my ($self, $recip, $append) = @_;
    my $file = $self->recip2name($recip);
    my $mode = $append ? '>>' : '<';
    my $dir  = dirname($file);
    my $fh;
    unless (open($fh, $mode, $file)) {
	eval {
	    -d $dir or mkpath($dir) or die "can't create dir $dir: $!";
	    open($fh, $mode, $file) or die "can't open ($mode) $file: $!";
	};
	if ( $@ ) {
	    warn "Error saving digest to cache: $@ - Skipping digest cache for this user";
            return undef;
	}
    }
    binmode($fh) if defined($fh);
    return $fh;
}

sub clear {
    my ($self, $recip) = @_;
    my $file = $self->recip2name($recip);
    unlink $file if -e $file; 

    my $dir;
    my $path;
    ($path, $dir, undef) = $self->get_path($recip);
    rmdir "$path/$dir"; # rmdir will delete empty directories or return false
}

# Adds a message entry to the recipient's database.
sub add_message_digest {
    my ($self, $recipient, $envelope_to, $msg) = @_;
    my $wfh = $self->load($recipient, 1);
    return unless ( defined($wfh) );

    require PureMessage::Message::ComputedFields;
    print $wfh join(DELIM,
	"i" . $msg->id,
	"d" . str2time($msg->minfo_value('Date')),
	"F" . ($msg->minfo_value('Envelope-From') || ""),
	"T" . $envelope_to,
	"p" . ($msg->minfo_value('Spam-Probability') || 0),
	"f" . _enc($msg->header('From')),
	"s" . _enc($msg->c_subject_utf8),
	"S" . _enc($msg->c_subject_body_utf8),
	(map { "r$_" } $msg->minfo_value('Reason')),
	"z" . $msg->message_size,
    ), "\n";

    close($wfh) or die "can't save $recipient cache: $!";
}

# Make values safe for dumping in a single line. The trick is, because we're
# only writing a digest, I don't *decode*. The encoded values are what get
# written to the digests.
sub _enc {
    local $_ = shift;
    return '' unless defined;
    s/\s+/ /g;
    s/\0//g;
    $_;
}

my %NAME = (
    i => 'id',
    d => 'time',
    p => 'spamprob',
    f => 'from',
    F => 'envfrom',
    T => 'envto',
    s => 'subject',
    S => 'subjbody',
    r => 'reasons',
    z => 'size',
);

# For each entry in the database, 
sub recipient_scan {
    my ($self, $cb) = @_;
    my $PATH2FILE = qr,.*/(.*),;
    my $wanted = sub {
	return unless -f $File::Find::name;
	return unless open(my $rfh, $File::Find::name); 

	(my $recip = $File::Find::name) =~ s/$PATH2FILE/$1/;
	my @msgs;
	while (my $digest = <$rfh>) {
	    chomp $digest;
	    my %info;
	    for my $info (split(DELIM, $digest)) {
		my $k = substr($info, 0, 1, '');
		unless (defined $NAME{$k}) {
		    warn "corrupt cache file $File::Find::name:$.: "
			. "unknown key '$k' skipped\n";
		    next;
		}
		if ($k eq 'r') {
		    push @{$info{$NAME{$k}}}, $info;
		}
		else {
		    $info{$NAME{$k}} = $info;
		}
	    }
	    push @msgs, \%info;
	}

	&$cb(fname2addr($recip), \@msgs);
    };
    find({ wanted => $wanted,
	   # Skip the dryrun subdirectory in for-real mode
	   preprocess => $self->{dryrun} ? () : sub {
	       return () if $_ eq DRYDIR;
	       @_;
	   },
	   no_chdir => 1,
    }, join "/", @{$self->{dir}});
}

1;
