package PureMessage::QDigest::DB;

use strict;
use Fcntl qw(O_RDWR O_CREAT);
use ActiveState::File::Atomic;
use PureMessage::Config qw(%config);

my $block_size = $config{block_size};

# These are required for the require/import section in the constructor
our $DB_HASH;
sub R_NEXT;
sub R_FIRST;

sub new {
    my ($class, $file) = @_;
    my $lockf = "$file.lock";
    
    eval {
        require DB_File;
        DB_File->import('$DB_HASH', 'R_FIRST', 'R_NEXT');  
    };
    if ($@) {
        die("Local quarantine digests are not supported!\n");
    }
    # Get an exclusive lock on the database. Print the database (if it
    # exists) to a temporary file, and then close the tempfile and open
    # *that* with DB_File. If anything goes wrong, we'll lose the changes,
    # but we won't corrupt the original file. You must call $self->save()
    # to commit the changes.
    my $lock = ActiveState::File::Atomic->new(
	$lockf, writable => 1, create => 1
    );
    my $temp = "$file.$$";
    if (-f $file && -s $file) {
        open (my $ORIG, $file);
	open (my $CP, '>', $temp) or die "can't write $temp: $!";
	binmode($ORIG);
	binmode($CP);
	local $/ = \4096;
	while (<$ORIG>) {
	    print $CP $_;
	}
	close $CP or die "can't write $temp: $!";
	close $ORIG;
    }

    my $db = DB_File->new($temp, O_RDWR|O_CREAT, 0666, $DB_HASH)
	or die "can't open qdigest database '$temp': $!";

    bless {
	db        => $db,
	_file     => $file,
	_lock     => $lock,
	_tmp_file => $temp,
    }, ref($class) || $class;
}

sub DESTROY {
    my $self = shift;
    delete $self->{db};
    $self->{_lock}->close if defined $self->{_lock};
    unlink $self->{_tmp_file} if defined $self->{_tmp_file};
}

sub save {
    my $self = shift;
    $self->{db}->sync;
    delete $self->{db};
    rename $self->{_tmp_file}, $self->{_file};
    delete $self->{_tmp_file};
    $self->{_lock}->close;
    delete $self->{_lock};
}

sub DELIM() { "\1" }
sub ANY()   { '@'  }

# Returns the minimum id to begin scanning from. If the optional
# PureMessage::List object is provided, the minimum ID from all of their
# digests is returned. Otherwise, the minimum ID of the magical ANY user is
# returned.
sub min_digest_id {
    my $self = shift;
    my $list = shift;
    my $reasons = shift;
    my $db = $self->{db};

    my %reasons_seen = map { $_ => 0 } @$reasons if $reasons;
    my $min;
    my ($key, $val, $st);
    for ($st = $db->seq($key, $val, R_FIRST);
	 $st == 0;
	 $st = $db->seq($key, $val, R_NEXT))
    {
	my ($user, $digest) = split(DELIM, $key);
	my ($id, $date) = split(DELIM, $val);

	# If they provided '$list', only consider users in $list
	if ($list) {
	    next unless $user ne ANY and $list->match($user);
	}
	# otherwise only consider the ANY user
	else {
	    next unless $user eq ANY;
            $reasons_seen{$digest}++
                if ($reasons && exists $reasons_seen{$digest});
	}
	$min = $id if not defined $min or $id < $min;
    }

    # If 'reasons' was specified make sure ALL the reasons have an entry
    # If they dont, return 0    
    unless($list || !%reasons_seen) {
        foreach my $reason (keys %reasons_seen) {
            return undef unless $reasons_seen{$reason};
        }
    }
    $min;
}

# Returns the last id digested for the given recipient/digest. If the digest
# isn't specified, the earliest id for the recipient is returned.
sub last_id {
    my ($self, $recip, @digests) = @_;
    $recip = ANY unless defined $recip;
    my $db = $self->{db};
    my $min;
    for my $d (@digests) {
	my ($key, $val);
	$key = join(DELIM, $recip, $d);
	next if $db->get($key, $val);
	my ($id, $date) = split(DELIM, $val);
	$min = $id if not defined $min or $id < $min;
    }
    $min;
}

# Returns the last date the digester ran for the given recipient/digest. If
# multiple digests are specified, the date of the earliest id for the
# recipient is returned.
sub last_date {
    my ($self, $recip, @digests) = @_;
    $recip = ANY unless defined $recip;
    my $db = $self->{db};
    my ($min, $mdate);
    for my $d (@digests) {
	my ($key, $val);
	$key = join(DELIM, $recip, $d);
	next if $db->get($key, $val);
	my ($id, $date) = split(DELIM, $val);
	if (not defined $min or $id < $min) {
	    $min = $id;
	    $mdate = $date;
	}
    }
    $mdate;
}

sub update_last_id {
    my ($self, $recip, $digest, $id) = @_;
    $recip = ANY unless defined $recip;
    my $key = join(DELIM, $recip, $digest);
    my $val;
    my $date = ''; # XXX okay?
    if ($self->{db}->get($key, $val) == 0) {
	my ($id, $d) = split(DELIM, $val);
	$date = $d;
    }
    $val = join(DELIM, $id, $date);
    $self->{db}->put($key, $val);
}

sub update_last_date {
    my ($self, $recip, $digest) = @_;
    $recip = ANY unless defined $recip;
    my $key = join(DELIM, $recip, $digest);
    my $val;
    my $id = 0;
    if ($self->{db}->get($key, $val) == 0) {
	my ($i, $date) = split(DELIM, $val);
	$id = $i;
    }
    $val = join(DELIM, $id, time);
    $self->{db}->put($key, $val);
}

# Scan the database
sub scan {
    my ($self, $cb) = @_;
    my ($key, $val, $st);
    my $db = $self->{db};
    for ($st = $db->seq($key, $val, R_FIRST);
	 $st == 0;
	 $st = $db->seq($key, $val, R_NEXT))
    {
	my ($user, $digest) = split(DELIM, $key);
	my ($last_id, $last_date) = split(DELIM, $val);
	&$cb($user, $digest, $last_id, $last_date);
    }
}

1;
