package PureMessage::QDigest::Scanner;

# This module encapsulates functionality for scanning the quarantine for
# message digesting purposes.

# Error handling: routines in this module (at least the external ones)
# should not die(). They should log the error and return status.

# Note that this is the first module in the PM code base where the
# centralized index (a.k.a. "pmdb quarantine") is actually referred to
# as index.

use base 'Exporter';
use strict;
use warnings;
use PureMessage::Logger;
use HTTP::Date qw(time2iso);
use PureMessage::List;

our @EXPORT_OK = qw(message_scan);

# Apply the specified callback to all the messages in the db that
# satisfy the selection criteria. Returns the count
# of rows scanned if all went well; undef if something went wrong.

# XXX There's too many parameters here. For now, we have to keep them
# positional for interface purposes, though
sub message_scan {

    # formerly known as "centralized quarantine"
    my $index = shift;

    my $callback = shift;
    my $folder = shift; # unused (placeholder)
    my $earliest = shift;
    my $latest = shift;
    my $reverse = shift; # unused (placeholder)
    my $reasons = shift;
    my $recipients = shift;
    my $override_state = shift;
    my $rows_scanned = 0;

    my $message;

    my $logger = PureMessage::Logger->new(ident => 'digest-scanner');

    eval {
        _prepare_scan('index'      => $index,
                      'earliest'   => $earliest,
                      'latest'     => $latest,
                      'logger'     => $logger,
                      'reasons'    => $reasons,
                      'recipients' => $recipients,
                      'override_state' => $override_state);
    };
    if ($@) {
        $logger->log("ERROR",
                     "Error encountered while preparing message scan: $@"
        );
        return undef;
    }

    eval {
        $rows_scanned = _do_scan($index, $callback, $logger);
    };
    if ($@) {
        $logger->log("ERROR",
                     "Error encountered while performing message scan: $@"
        );
        return undef;
    }

    return $rows_scanned;
}

# DB interaction => needs to be eval-ed.
sub _prepare_scan {
    my %arg = @_;
    my $index = $arg{'index'};
    my $earliest = $arg{'earliest'};
    my $latest = $arg{'latest'};
    my $logger = $arg{'logger'};
    my $quarantine_reasons = $arg{'reasons'};
    my $recipients = $arg{'recipients'};
    my $override_state = $arg{'override_state'};

    my $approve_numeric = $index->action_numeric('approve');
    my $delete_numeric = $index->action_numeric('delete');
    
    my $dbh = $index->{pmxdir_dbh};

    # message/recipient combinations for non-approved, non-deleted
    # messages that arrived after the most recent digested message for
    # the recipient
    my $extract_digest_sql;
    my $table_exists = 0;
    {
        my $hStmt = $dbh->prepare("select count(*) from pg_class where relname = 'temp_recipient_message_digest'");
        $hStmt->execute;
        my @row = $hStmt->fetchrow_array;
        if (1 == $row[0]) { 
            $table_exists = 1; 
        }
    }        
    
    # Table is mis-named. It is a permanent table, not a temp table.
    # 'pmx-qdigest --send-only' option reads from the related view 'view_temp_digest_state'.
    if ($table_exists) {
        # TRUNCATE instead of DELETE to avoid a table scan 
        eval { $dbh->do("truncate table temp_recipient_message_digest") };
        if ($@) { $dbh->rollback(); }
    
        $logger->log("DEBUG", "Inserting new data into an existing, emptied table");
        $extract_digest_sql = 'insert into temp_recipient_message_digest ';

    } else {
        $logger->log("DEBUG", "Initial recipient message digest table creation");
        $extract_digest_sql = 'create table temp_recipient_message_digest as ';
    }   

    # Using the same select stmt for both table creation and insertion.
    $extract_digest_sql .=<<EOT;
select message.m_global_id as global_id,
       h_from_local || '\@' || h_from_domain as header_from,
       env_from_local || '\@' || env_from_domain as envelope_from,
       message.c_subject_utf8,
       message.c_subject_body_utf8,
       message.m_date, m_reason.name as reason_name,
       message.m_message_size as size,
       message.m_spam_probability as spam_probability,
       recipient.local || '\@' || recipient.domain as recipient,
       recipient.id as recipient_id
from message, message_recipient, recipient, m_reason
where message.id = message_recipient.message_id
and message_recipient.recipient_id = recipient.id
and message.reason_id = m_reason.reason_id
and message_recipient.is_envelope = 1
and (
        (message_recipient.action is null)
        or
        (message_recipient.action not in ($approve_numeric, $delete_numeric))
)
EOT
    my $start_scan_predicate = _start_scan_predicate($earliest,
                                                     $override_state);
    if ($start_scan_predicate) {
        $extract_digest_sql .= " and $start_scan_predicate";
    }
    my $end_scan_predicate = _end_scan_predicate($latest);
    if ($end_scan_predicate) {
        $extract_digest_sql .= " and $end_scan_predicate";
    }
    my $reason_predicate = _reason_predicate($quarantine_reasons);
    if ($reason_predicate) {
        $extract_digest_sql .= " and $reason_predicate";
    }

    $logger->log("DEBUG", "extract_digest_sql: $extract_digest_sql");
    $dbh->do($extract_digest_sql);

    if (! $table_exists) {
        $logger->log("DEBUG", "Creating index and view on 'temp' digest table ...");
        $dbh->do(<<EOT);
create index idx_tmp_rcpnt_dgst_rcpnt_rsn
on temp_recipient_message_digest(recipient, reason_name)
EOT
        $dbh->do(<<EOT);
create view view_temp_digest_state as
select recipient, reason_name, max(m_date) as last_m_date
from temp_recipient_message_digest group by recipient, reason_name
EOT
    }
    
    $dbh->commit();
    $dbh->{AutoCommit} = 1;
    $dbh->do("analyze temp_recipient_message_digest");
    $dbh->{AutoCommit} = 0;

    # now eliminate recipient/message combinations for the recipients
    # that don't match the specified recipients list, if needed;
    # we do this "the hard way" to avoid the need to translate list/email
    # to SQL pattern matching
    if ($recipients) {
        $logger->log("DEBUG", "Eliminating unwanted recipients...");
        my $recipient_scan_stmt = $dbh->prepare(<<EOT);
select distinct recipient from temp_recipient_message_digest
EOT
        $recipient_scan_stmt->execute();
        my $recipient;
        $recipient_scan_stmt->bind_columns(\$recipient);
        my $delete_for_recipient_stmt = $dbh->prepare(<<EOT);
delete from temp_recipient_message_digest where recipient = ?
EOT
        while ($recipient_scan_stmt->fetch()) {
            unless ($recipients->match($recipient)) {
                $delete_for_recipient_stmt->execute($recipient);
            }
        }
        $dbh->commit();
        $dbh->{AutoCommit} = 1;
        $dbh->do("analyze temp_recipient_message_digest");
        $dbh->{AutoCommit} = 0;
    }
}

# db interaction => needs to be eval-ed
# we're scanning data in batches here to avoid PGSQL's client side
# buffering issues.
# this can cause that the same message is scanned more than once (if its
# rows are split across two batches), but that's ok, since we're actually
# scanning message/recipient pairs anyway.
# returns the count of scanned rows.
sub _do_scan {
    my $index = shift;
    my $callback = shift;
    my $logger = shift;
    my $message_count = 0;
    my $get_msg_data_sql =<<EOT;
select global_id, c_subject_utf8, c_subject_body_utf8,
       header_from, envelope_from, spam_probability,
       m_date, reason_name, size, recipient
from temp_recipient_message_digest
order by global_id, m_date
EOT

    # env var is mainly here for testing purposes...
    my $batch_size = $ENV{'PMX_TEST_SCAN_BATCH_SIZE'} || 100000;
    my $offset = 0;
    my $rows_retrieved = 0;
    my $offset_increment = 0;
    do {
        $offset += $offset_increment;
        my $batch_sql = "$get_msg_data_sql " .
                        "offset $offset limit $batch_size";
        $offset_increment = _do_scan_batch('sql'              => $batch_sql,
                                           'callback'         => $callback,
                                           'index'            => $index,
                                           'logger'           => $logger);
        $rows_retrieved += $offset_increment;
    } until ($offset_increment < $batch_size);
    return $rows_retrieved;
}

# scans one batch of digest data, using the limit/offset facility
# returns the count of rows retrieved (so that we can calculate offset
# next time)
sub _do_scan_batch {
    my %opt = @_;
    my $get_msg_data_sql = $opt{sql};
    my $callback = $opt{callback};
    my $index = $opt{index};
    my $logger = $opt{logger};
    my ($m_global_id, $c_subject_body_utf8, $c_subject_utf8, $m_date,
        $header_from, $envelope_from, $spam_probability,
        $m_reason, $m_size, $m_envelope_to);
    my $dbh = $index->{pmxdir_dbh};
    my $stmt = $dbh->prepare($get_msg_data_sql);
    $logger->log("DEBUG", "batch scan SQL: $get_msg_data_sql");
    $stmt->execute();
    $stmt->bind_columns(\$m_global_id, \$c_subject_utf8, \$c_subject_body_utf8,
                        \$header_from, \$envelope_from, \$spam_probability,
                        \$m_date, \$m_reason, \$m_size, \$m_envelope_to);
    my $current_id = "";
    my $message;
    my $row_count = 0;
    my $more_data = 1;
    do {
        if ($more_data = $stmt->fetch()) {
            $row_count ++;
        }

        my $do_construct_message =
            (!$message || ($current_id ne $m_global_id)) && $more_data;
        my $do_append_recipient =
            $message && ($current_id eq $m_global_id) && $more_data;
        my $do_process_message =
            $message && (($current_id ne $m_global_id) || (!$more_data));

        if ($do_process_message && $callback) {
            &$callback($message);
        }
        if ($do_construct_message) {
            $message = $index->message_lite(
                {id               => $m_global_id,
                 subject          => $c_subject_utf8,
                 subject_body     => $c_subject_body_utf8,
                 header_from      => $header_from,
                 envelope_from    => $envelope_from,
                 spam_probability => $spam_probability,
                 m_date           => $m_date,
                 reason           => $m_reason,
                 size             => $m_size,
                 recipient        => $m_envelope_to}
            );
            if ($message) {
                $current_id = $m_global_id;
            }
            else {
                $logger->log("WARNING",
                             "Unable to construct message $m_global_id");
            }
        }
        if ($do_append_recipient) {
            $message->append_envelope_recipient($m_envelope_to);
        }
    } until (!$more_data);
    return $row_count;
}

sub _start_scan_predicate {
    my $earliest = shift;
    my $override_state = shift;
    my $predicate = "(1 = 1)";
    my $state_predicate = <<EOT;
(
        (message.m_date > (
            select last_m_date
            from digest_state
            where recipient.local || '\@' || recipient.domain = digest_user
            and m_reason.name = digest_type
        )
    )
    or (
        not exists (
            select 'x' from digest_state
            where recipient.local || '\@' || recipient.domain = digest_user
            and m_reason.name = digest_type
        )
    )
)
EOT
    if ($earliest) {
        $earliest = time2iso($earliest);
        $predicate = "('$earliest' < message.m_date)";
    }

    return $override_state ? "$predicate" : "($state_predicate and $predicate)";
}

sub _end_scan_predicate {
    my $predicate = "";
    my $latest = shift;
    if ($latest) {
        $latest = time2iso($latest);
        $predicate .= "(message.m_date <= '$latest')";
    }
    return $predicate;
}

sub _time_range_predicate {
    my $predicate = "";
    my $earliest = shift;
    my $latest = shift;
    $earliest = time2iso($earliest) if $earliest;
    $latest = time2iso($latest) if $latest;
    $predicate = " ('$earliest' < message.m_date)" if $earliest;
    $predicate .= " and (message.m_date <= '$latest') " if $latest;
    $predicate =~ s/^ and// if $predicate;
    return $predicate;
}

sub _reason_predicate {
    my $reasons = shift;

    my $predicate = "";
    if ($reasons) {
        $predicate = "(1=0)";
        if (scalar @$reasons) {
            # quarantine reasons are stored in lowercase in the db
            my @quoted_lowercase_reason = map {"'" . lc($_) .  "'"} @$reasons;
            $predicate = "m_reason.name in (" . join(', ', @quoted_lowercase_reason) . ")";
        }
    }
    return $predicate;
}

return 1;
