package PureMessage::QMan;
use base 'Exporter';

use strict;
use Mail::Address;
use Net::SMTP;
use PureMessage::Config qw(%config);
use PureMessage::Queue qw(queue_mail);
use PureMessage::MessageLog qw(join_line);
use Carp;
use LWP::UserAgent;
use Storable qw (thaw);

our $VERSION = '4.600';
our @EXPORT_OK = qw(approve forward save_to_mbox trash process_qaction
                    prepare_msg_preview fetch_attachment
                    full_clean_body full_message_source full_attachments_info
                    full_message_minfo
                   );

sub trash {
    my $messages = shift;
    my %opts = @_;
    my $to = delete $opts{to} if exists $opts{to};
    $messages = [$messages] unless ref($messages) && ref($messages) eq 'ARRAY';
    my $logger = $opts{logger} ||
	croak "'trash()' requires a logger parameter";

    my $i = 0;

    # XXX If this is done from a web interface it may timeout if there are
    # several messages to trash and some of them are on remote hosts. The
    # remote trash function could be taken care of by a background process.

    foreach my $message (@{$messages}) {
	my $id = $message->id;
	eval {

	    # If not all the recipients are trashing this message then don't
	    # remove it. Just mark it as trashed for those recipients. 
	    # Otherwise delete the whole message in all locations.
        # XXX GroupUI needs the status can be kept between the interval
        # of running of pmx-qrelease and pmx-qmeta-index, so we must do
        # both here
        if ($opts{tag_and_trash}) {
            _record_action('delete', $message, $to);
            _full_trash($message, $logger);
            $logger->log("INFO", "Message $id has been tagged and trashed for deletion");
        }
	    elsif ($to) {
		$to = $opts{mark_user} if $opts{mark_user};
		_record_action('delete', $message, $to);
		$logger->log("INFO", 
		    "Marked message $id as deleted for one or more users");
	    }
	    else {
		_full_trash($message, $logger);
	    }
	    $opts{on_trashed}->($message) if exists $opts{on_trashed};
	    $i++;
	};
	if ($@) {
	    $logger->log("ERROR", "Failed to trash the message $id: $@");
	}
    }
    return $i;
}

sub forward {
    my ($messages, %opts) = @_;
    my $recip = $opts{to};
    my $logger = $opts{logger} ||
	croak "'forward()' requires a logger parameter";
    
    $messages = [$messages] unless ref($messages) && ref($messages) eq 'ARRAY';

    my $i = 0;
    foreach my $msg (@$messages) {
	my $id = $msg->id;
        eval {
	    if ($msg->is_local) {
		_qmail($msg, $config{admin}, to => $recip, %opts);
		$logger->log("INFO", "Forwarded message $id");
	    }
	    elsif ($config{async_release}) {
                unless (_is_centralized_quarantine()) {
		    croak "async_release configured with " .
                          "non-centralized quarantine";
                }
                $opts{request_type} = 'forward';
		_submit_release_request($id, %opts);
	    }
	    else {
		my $rq = _remote_quarantine();
		$recip = [$msg->minfo_value("Envelope-To")] unless $recip;
		$rq->forward($msg, $config{admin}, $recip, %opts);
	    }
	};
        if ($@) {
	    $logger->log("ERROR", "Error while forwarding message $id: $@");
        } else {
	    $i++;
        }
    }
    return $i;
}

sub approve {
    my ($messages, %opts) = @_;
    my $msgs = ref($messages) eq "ARRAY" ? $messages : [$messages];
    my $logger = $opts{logger} 
	|| croak "'approve()' requires a logger parameter";
    my $i = 0;
    $opts{action} = 'approve';
    foreach my $msg (@$msgs) {
	my $id = $msg->id;
	eval {
	    my $to;
	    if ($msg->is_local) {
		$to = _local_approve($msg, %opts);
	    }
            elsif ($config{async_release}) {
                unless (_is_centralized_quarantine()) {
		    croak "async_release configured with " .
                          "non-centralized quarantine";
                }
                $opts{request_type} = 'approve';
		$to = _submit_release_request($id, %opts);
                
                # normally we only exec the on_sent when we actually enqueue the
                # message for delivery, but for remote release we will consider
                # it enqueued when we submit the message to approve to the db 
                if (exists $opts{on_sent}) {
                    die 'parameter "on_sent" must be a subroutine',
                        unless (ref($opts{on_sent}) eq 'CODE');
                    $opts{on_sent}->($id, $msg, $to);
                }

                
            }
	    else {
		my $rq = _remote_quarantine();
		if ($to = $rq->approve($msg, %opts)) {
		    my $loc = $msg->locations->[0];
		    $logger->log("INFO", 
			"Dispatched approval for message $id to $loc");
		}
	    }
	    $to = $opts{mark_user} if $opts{mark_user};
            # record an action if it's an array ref with values or a scalar
	    _record_action("approve", $msg, $to) if (ref($to) eq 'ARRAY' and scalar @$to) or (!ref($to) and $to);
	};
	if ($@) {
	    $logger->log("ERROR", "Error while sending approved message $id: $@");
        } else {
	    $i++;
        }
    }
    return $i;
}

sub _local_approve {
    my ($msg, %opts) = @_;
    my $from_addr = $config{quarantine_preserve_env_from} ? $msg->minfo_value('Envelope-From') : $config{admin};

    my $to = _qmail($msg, $from_addr, trash => 0, %opts);
    _log_approval($msg, to => $to, %opts);
    return $to;
}


sub _full_trash {
    my ($msg, $logger) = @_;

    my $id = $msg->id;

    # If a message is local, just trash it and return.
    # Otherwise, submit the remote delete request either into the central
    # request queue or via rpc to the appropriate mail processing server.
    if ($msg->is_local) {
	$msg->trash;
	$logger->log("DEBUG", "Trashed local message $id");
	return;
    }
    elsif ($config{async_release}) {
        unless (_is_centralized_quarantine()) {
            croak "async_release configured with " .
                  "non-centralized quarantine";
        }
        _submit_delete_request($msg);
    }
    # this is legacy RPC stuff
    else {
        my @dsp_hosts = @{$msg->locations};
        if (@dsp_hosts) {
            my $rq = _remote_quarantine();
            $rq->trash($msg, $logger, @dsp_hosts);
        }
    }
}

sub save_to_mbox {
    my $messages = shift;
    my %opts = @_;
    my $file = $opts{mbox} ||
	croak "'mbox' is a required parameter when calling save_to_mbox()";
    my $append = $opts{mbox_append};
    my $MBOX;

    my $opened;
    if (ref(\$file) eq 'GLOB' or ref($file)) {
        $MBOX = $file;
    }
    else {
        my $mode = $append ? ">>" : ">";
        open($MBOX, $mode, $file) or die ("Failed to open '$file': $!");
	$opened++;
    }

    for my $msg (@{$messages}) {
        prepare_msg_preview($msg);
        my $from = $msg->minfo_value('Envelope-From') || "unknown";
        my $from_line = "From $from " . localtime();
        $from_line =~ s/\n//g;
        print $MBOX $from_line."\n";
        my $buf;
        my $leftover = "";
        my $end_nl;
        while($msg->read($buf,4*1024)) {
            # MBOX encoding
            # XXX Added another > to ">From" is not standard,
            # XXX see http://wp.netscape.com/eng/mozilla/2.0/relnotes/demo/content-length.html
            if (length $leftover) {
                substr($buf, 0, 0) = $leftover;
                $leftover = "";
            }
            if ($buf =~ s/^([F>].*)\z//m) {
                $leftover = $1;
            }
            $buf =~ s/^(>*From) />$1 /mg;
            # write buffer
            $buf =~ s/\r//g;
            print $MBOX $buf if length $buf;
            $end_nl = $buf =~ /\n\z/;
        }
        # write blank line
        if ($leftover) {
            $leftover =~ s/\r//g;
            print $MBOX $leftover;
            $end_nl = 0;
        }
        print $MBOX "\n" unless $end_nl;
        print $MBOX "\n";
        $opts{on_saved}->($msg) if exists $opts{on_saved};
    }
    close $MBOX if $opened;
    return scalar @{$messages};
}

sub _record_action {
    my ($action, $msg, $recip, $negate) = @_;
    return $msg->user_actions($recip, $action, $negate);
}

sub _qmail {
    # single message 
    my $msg = shift;

    # from_addr - same for each message
    my $from_addr = shift;

    my %opts = @_;

    # to_addr or array of to_addrs - destination(s) for each message
    my $to_addr = $opts{to};
    if (ref $to_addr and ref $to_addr ne 'ARRAY') {
	my $r = ref $to_addr;
	die "parameter 'to_addr' must be a scalar or an arrayref, not a '$r'";
    }

    my $error;
    if (exists $opts{on_error}) {
        die 'parameter "on_error" must be a subroutine',
            unless (ref($opts{on_error}) eq 'CODE');
        $error = sub {$opts{on_error}->(@_); return 0};
    }
    else {
        $error = sub { die $_[2] };
    }

    if (exists $opts{on_sent}) {
        die 'parameter "on_sent" must be a subroutine',
            unless (ref($opts{on_sent}) eq 'CODE');
    }

    local $Mail::Address::VERBOSE = 0; # silence "Unmatched () ..." warnings
    my @addrs = Mail::Address->parse($from_addr);
    return $error->($msg->id,'bad_sender_address',"Unable to parse sender address '$from_addr' $!") if scalar @addrs < 1;
    my $from = $addrs[0]->address();

    my $id = $msg->id;

    # get the to address from the parameters or from the envelope
    # information
    my $to;
    if ($to_addr) {
	$to = (ref $to_addr) ? $to_addr : [$to_addr];
    }
    else {
	$to = [$msg->minfo_value("Envelope-To")];
    }

    # Add a copy of the local message into the mail queue
    my $l_msg = $msg;
    if ($msg->can("get_local_message")) {
	$l_msg = $msg->get_local_message;
    }
    queue_mail(from => $from, to => $to, message => $l_msg);

    $opts{on_sent}->($id, $msg, $to) if exists $opts{on_sent};
    $msg->trash if $opts{trash};
    return $to;
}

# submits async delete request(s) for the specified message into the
# centralized queue; returns true on success; false otherwise
sub _submit_delete_request {
    my $msg = shift;
    my $store = $msg->message_store();
    my $logger = $store->{pmxdir_logger};
    my $action_queue = $store->get_release_queue();
    my $id = $msg->id();
    my $rqst_info = "delete request for message: $id";
    if ($action_queue->submit_request(global_id => $id,
                                      request_type => 'delete')) {
        $logger->log("DEBUG", "Submitted $rqst_info");
        return 1;
    } else {
	$logger->log("WARNING", "Unable to submit $rqst_info");
        return 0;
    }
}

# submits async release request(s) for the specified message into the
# centralized queue; returns reference to the array of recipients
sub _submit_release_request {
    my ($id, %opts) = @_;
    my $to_addr = $opts{to};
    my $to = (ref $to_addr) ? $to_addr : [$to_addr];
    my $store = $opts{quarantine} ||
        croak("Invalid 'quarantine' option, " .
              "unable to submit release request");
    my $logger = $opts{logger} ||
	croak "'_submit_release_request()' requires a logger parameter";
    my $async_queue = $store->get_release_queue();
    foreach my $addr (@$to) {
        my $rqst_info = "release request for message: $id, recipient: $addr";
        my ($local, $domain) = $store->split_address($addr);
        my $request_type = $opts{request_type};
        if ($async_queue->submit_request(global_id => $id,
                                         request_type => $request_type,
                                         local => $local,
                                         domain => $domain)) {
            $logger->log("DEBUG", "Submitted $rqst_info");
        } else {
	    $logger->log("WARNING", "Unable to submit $rqst_info");
        }
    }
    return $to;
}

# Processes quarantine action requests enqueued in the central db.
# Returns the count of processed requests. Deletes successfully
# enqueued requests from the centralized storage.
sub process_qaction {
    my $store = shift;
    my $max_attempt_count = shift;
    my $verbose = shift || 0;
    my $count_processed = 0;
    my $location = $store->{pmxdir_location};
    my $async_queue = $store->get_release_queue(verbose => $verbose);
    my $requests_ref = $async_queue->get_requests($location);
    my @request_to_delete;
    my @request_to_increment;
    my $logger = $store->{pmxdir_logger} ||
	croak "'process_qaction()' requires a logger parameter";
    my $global_id;
    foreach my $request_ref (@$requests_ref) {
        my $processed = 0;

        # we don't want to die in this loop...
        eval {
            return unless ($request_ref->{request_type});
            if (($request_ref->{request_type} eq 'approve') or
                ($request_ref->{request_type} eq 'forward')) {
                $processed = _process_release_request($store,
                                                      $request_ref,
                                                      $verbose);
            }
            elsif ($request_ref->{request_type} eq 'delete') {
                $processed = _process_delete_request($store,
                                                     $request_ref,
                                                     $verbose);
            }
        };
        if ($processed and (!$@)) {

             # everything went fine
             push @request_to_delete, $request_ref;
             $count_processed++;
        }
        else {
            my $message = "";
            $message .= $@ if $@;
            grep($request_ref->{request_type}, (qw(approve forward delete)))
                or ($message .= "Invalid request");

            _handle_failed_request($logger, $request_ref, $max_attempt_count,
                                   \@request_to_delete, \@request_to_increment,
                                   $message);
        }
    }

    if (scalar @request_to_delete) {
        unless ((scalar @request_to_delete) == 
                 $async_queue->delete_requests(\@request_to_delete)) {
	    $logger->log("ERROR",
                         "Error while removing quarantine requests");
        }
    }
    if (scalar @request_to_increment) {
        unless ((scalar @request_to_increment) == 
                 $async_queue->inc_attempt_counts(\@request_to_increment)) {
	    $logger->log("ERROR",
                         "Error while incrementing attempt counts");
        }
    }
    return $count_processed;
}

# inserts message body data into the database (if needed)
# returns true if the insert succeeds (or the data was previously
# inserted), otherwise returns false.
my $user_agent = undef;
sub prepare_msg_preview {
    my $message = shift;
    my $message_store = $message->{pmxdir_store};
    my $data_prepared = 1;
    my $insert_body = !$message_store->is_message_body_inserted($message->id());
    my $insert_hits = 0; 
    if ($insert_body or $insert_hits) {

        # get the location from the message, map it to the URL of the
        # CGI that'll insert the message body, and invoke that CGI
        my $location_name = ${$message->locations()}[0];
        require PureMessage::Resources::Location;
        my $location;
        if ($location = PureMessage::Resources::Location->Find($location_name)) {
            my $hostname = $location->hostname();
            my $protocol = $location->manager_protocol();
            my $port     = $location->manager_port();
            my $global_id = $message->id();
            require LWP::UserAgent;
            $user_agent ||= LWP::UserAgent->new();
            my $request = "$protocol://$hostname:$port/unauthenticated/pmx-qmeta-index.cgi?global_id=$global_id&insert_body=$insert_body&insert_hits=$insert_hits";
            my $response = $user_agent->get($request);
            if ($response->is_success) {
                my $content = $response->content();
                chomp $content;
                my ($status, $remote_error) = split(/\|/, $content);
                unless (lc($status) eq "success") {

                    # application error
                    $remote_error = "" unless $remote_error;
                    $message_store->log(
                        "ERROR",
                        "Request $request failed: $remote_error"
                    );
                    $data_prepared = 0;
                }
            }
            else {

                # protocol error
                my $error_text = $response->error_as_HTML;
                $error_text =~ s/<[^>]*>//gs;
                $message_store->log(
                    "ERROR",
                    "Request $request failed: " . $error_text
                );
                $data_prepared = 0;
            }
        }
        else {
            $message_store->log("ERROR", "Unable to access attributes of the location $location_name");
            $data_prepared = 0;
        }
    }
    return $data_prepared;
}

sub _handle_failed_request {
    my ($logger, $request_ref,
        $max_attempt_count, $deleted_ref, $incremented_ref, $message) = @_;
    my @message;
    push @message, $message if $message;

    if ($max_attempt_count <= (1 + $request_ref->{attempt_count})) {
        push @message, "request will be deleted";
        push @$deleted_ref, $request_ref;
    }
    else {
        push @message, "request will be retried";
        push @$incremented_ref, $request_ref;
    }
    _log_request_error($logger, $request_ref, join(', ', $message));
}

# returns true on success, false on failure
sub _process_delete_request {
    my $store = shift;
    my $logger = $store->{pmxdir_logger};
    my $request_ref = shift;
    my $verbose = shift;
    my $global_id = $request_ref->{global_id};
    my $id = $store->local_from_global($global_id);
    unless ($id) {
        _log_request_error($store->{pmxdir_logger},
                           $request_ref, "Invalid global id");
        return 0;
    }
    my $message;
    unless ($message = $store->message($id)) {
        $logger->log("WARNING", "Unable to initialize message $id");
        return 0;
    }
    print "Deleting message $global_id.\n" if $verbose;

    # we don't want any surprises in this call, therefore eval it...
    eval {$message->trash};
    if ($@) {
        $logger->log("WARNING", "Error while trashing message $id: $@");
    }
    else {
        $logger->log("DEBUG", "Trashed local message $id");
    }
    return 1;
}

# returns true on success, false on failure
sub _process_release_request {
    my $store = shift;
    my $request_ref = shift;
    my $verbose = shift;
    my $global_id = $request_ref->{global_id};
    my $id = $store->local_from_global($global_id);
    unless ($id) {
        _log_request_error($store->{pmxdir_logger},
                           $request_ref, "Invalid global id");
        return 0;
    }
    my $message = $store->message($id);
    my $recipient = $request_ref->{local};
    my $domain    = $request_ref->{domain};
    $recipient .= "\@" . $domain if $domain;
    print "Enqueuing message $id" . ($recipient ? " for $recipient" : " for delivery") . ".\n" if $verbose;
    my $to = _qmail($message, $config{admin}, to => $recipient);
    # Log approval details just for 'approved' messages.
    if ($request_ref->{request_type} eq 'approve') {
	_log_approval($message, to => $recipient, 
		      logger => $store->{pmxdir_logger});
    }
    return $to;
}

sub _log_request_error {
    my $logger = shift;
    my $request_ref = shift;
    my $error_info = shift;
    my $error_msg = "Error encountered while processing request ";
    $error_msg .= "global id: $request_ref->{global_id} ";
    $error_msg .= "request type: " . ($request_ref->{request_type} || "");
    $error_msg .= "local: $request_ref->{local} " if $request_ref->{local};
    $error_msg .= "domain: $request_ref->{domain} " if $request_ref->{domain};
    $error_msg .= ": $error_info" if $error_info;
    $logger->log("ERROR", $error_msg);
}

sub _log_approval {
    my ($messages, %opts) = @_;
    my $to = $opts{to};
    my $log = $opts{logger} || return;
    $messages = [$messages] unless ref($messages) eq 'ARRAY';
    foreach my $msg (@$messages) {
	$log->log("INFO", 
	    "Approval Details:" . _encode_approval_line($msg, $to));
    }
}

sub _encode_approval_line {
    my ($m, $to) = @_;
    my @attr = (id => $m->id);
    if ($to) {
	if (ref($to)) {
	    push(@attr, t => $_) foreach @$to;
	}
	else {
	    push(@attr, t => $to);
	}
    }

    my %a = qw(f Envelope-From p Spam-Probability reason Reason q Queue-Id);
    while (my ($k, $v) = each(%a)) {
	push(@attr, $k, $m->minfo_value($v)) if $m->minfo_value($v);
    }
    if ($m->minfo_value("Relay")) {
	(my $r = $m->minfo_value("Relay")) =~ s/.*\[([\d\.]+)\].*/$1/;
	push(@attr, 'r', $r) if $r;
    }
    push(@attr, ("h", $_)) foreach $m->minfo_value("spam-hit");
    return join_line(undef, @attr);
}

# perhaps "centralized index" would be a better wording
sub _is_centralized_quarantine {
    return ('pmdb' eq $config{quarantine_type});
}

sub _remote_quarantine {
    require PureMessage::QMan::RPCClient;
    return PureMessage::QMan::RPCClient->new();
}

sub full_message_minfo {
    my $msg = shift;
    my $store = $msg->message_store();
    my $logger = $store->{pmxdir_logger};
    my $serialized =  _remote_service_fetch($msg, 'minfo');
    my @minfo;
    eval {
        @minfo = @{ thaw $serialized };
    };
    if ($@) {
        $logger->log("WARNING", "Unable to fetch minfo: $@");
    }
    return \@minfo;
}

sub full_clean_body {
    my $msg = shift;
    return _remote_service_fetch($msg, 'message');
}

sub full_message_source {
    my $msg = shift;
    return _remote_service_fetch($msg, 'source');
}

sub full_attachments_info {
    # $msg is the message global_id.
    my $msg = shift;
    my $string = _remote_service_fetch($msg, 'attach_info');

    my $hashref = {};
  
    foreach (split "\n", $string) {
        my ($id, $name, $type, $size, $m_filename, $m_fullfilename, $m_filetype, $m_size, $m_text) = split(/\|/, $_,9);
	
        $hashref->{$id}{id}         = $id;
        $hashref->{$id}{filename}   = $name;
        $hashref->{$id}{type}       = $type;
        $hashref->{$id}{size}       = $size;

        my $minfo   = {};
        $minfo->{m_filename}     = $m_filename;
        $minfo->{m_fullfilename} = $m_fullfilename;
        $minfo->{m_filetype}     = $m_filetype;
        $minfo->{m_size}         = $m_size;
        $minfo->{m_text}         = $m_text;

        if ($m_size) {
            $hashref->{$id}{has_match}  = 1;
        }

        $hashref->{$id}{minfo}    ||= [ ];
        
        push(@{$hashref->{$id}{minfo}}, $minfo);
    }

    # parse into hashref
    return $hashref;
}

sub _remote_service_fetch {
    my ($msg, $service, %opts) = @_;

    my $global_id = $msg->minfo_value('Global-Id');
    my $location_name = ${$msg->locations}[0];
    die "Cannot fetch location name for message with global ID: $global_id\n" unless $location_name;

    require PureMessage::Resources::Location;
    my $location = PureMessage::Resources::Location->Find($location_name);
    die "Cannot get location for location name: $location_name\n" unless $location;
    my $hostname = $location->hostname;
    die "Cannot get hostname for message with global ID: $global_id\n" unless $hostname;

    my $base_url = "https://$hostname:28443";

    my $ua = LWP::UserAgent->new;
    my $url = "$base_url/services/quarantine/$service/$global_id";
    $url .= "/$opts{part_id}" if $opts{part_id};
    my $resp = $ua->request(HTTP::Request->new(GET => $url), $opts{callback});
    unless ($resp->is_success) {
        die "Couldn't fetch $url: " . $resp->status_line . "\n";
    }
    return $resp->content unless $opts{callback};
}

sub fetch_attachment {
    my ($msg, $part_id, $cb) = @_;
    _remote_service_fetch($msg, 'attachment', 
        part_id => $part_id, 
        callback => $cb);
}

1;

__END__

=head1 NAME

PureMessage::QMan - Quarantine management functions

=head1 SYNPOSIS

  use PureMessage::QMan qw(approve forward save_to_mbox trash);

  approve($messages, to=>'me@here.com', %sendmail_opts)

  forward($messages, $to, %sendmail_opts)

  save_to_mbox($messages, mbox => $file, mbox_append => $append) 

  trash($messages)

  * $messages - can be both an arrayref or a single PureMessage::Message
  * every function returns the number of successfully
    approved/forwarded/saved/etc messages
  * multipart=>1 means to send multiple messages as one multipart MIME
    message.

=head1 DESCRIPTION

This module provide functions for quarantine management.  The
following functions are provided:

=over

=item approve( $messages, %opts )

This function approves the given messages.  This will send
quarantined messages to the original recipients.

=item forward( $messages, $recipients, %opts )

This function will send the given messages to given recipients.

=item save_to_mbox( $messages, mbox => $file, mbox_append => $append )

This function will extract save the given set of messages to the file
$file in mbox format.  If the $append argument is TRUE, then the
messages will be appended to the given $file.

=item trash( $messages )

This function will trash the given messages.

=item prepare_msg_preview($message)

Enables message preview by inserting the message body (or part of it)
into the centralized storage, accessible by the previewing functionality.
Returns true if the insert succeeds (or the data was previously
inserted), otherwise returns false.

=item full_clean_body($message)

Uses LWP to fetch the full cleaned message body directly from the 
edge server on which it resides.

=item full_message_headers($message)

Uses LWP to fetch the full message headers directly from the 
edge server on which it resides.

=item full_attachments_info($message)

Uses LWP to fetch the full attachment info directly from the edge server on
which it resides.

=item fetch_attachment($message, $part_id, $callback)

Use LWP to fetch a message attachment identified by $part_id from an edge
server, calling $callback for each chunk of data.

=back

=head1 SEE ALSO

L<PureMessage::MessageStore>

=cut
