package PureMessage::QMan::RPCClient;
use strict;
use warnings;

use PureMessage::RPC::ClientV2;
use PureMessage::ServerGroup;

sub new {
    my $p = shift;
    my $class = ref($p) || $p;
    return bless {}, $class;
}

# XXX It would be a lot better if this could process more than one message
# at a time and could send them in batches to the remote host. Connections
# could be cached as well.
sub approve {
    my ($self, $msg, %opts) = @_;
    my $msgs = ref($msg) ? $msg : [$msg];

    # Can't pass these options over the wire
    my $logger = delete $opts{logger};
    delete $opts{on_sent};
    delete $opts{on_error};

    # XXX We can only send scalar values within the options hash so pack the
    # list of addresses so that they can be sent as a scalar and unpacked
    # on the other side.
    $opts{to} = "\0" . join("\0", @{$opts{to}})
	if ref($opts{to}) eq "ARRAY";
	
    # Only a single location needs to do the approval (or the reciptient(s)
    # would get multiple coppies). For now we always choose the first one.
    my $hname = $msg->locations->[0];
   
    # Server groups records the rpc/sync service port and proto under
    # the sync_* variables. Map these on top of 'proto' and 'port' as the 
    # rpc client ignores the sync_* options.
    my $host = PureMessage::ServerGroup::get_host($hname);
    $host->{port} = $host->{sync_port};
    $host->{proto} = $host->{sync_proto};
    my $rpc = PureMessage::RPC::ClientV2->new({$hname => $host});

    # Get the message's local ID and make the request.
    my $id = $msg->local_id($hname);
    $rpc->call("QMan.approve", msgid => $id, %opts);
    $rpc->wait();
    if (my $err = $rpc->error($hname)) {
	$logger->log("ERROR", "Failed to dispatch approval for message $id to $hname: $err") 
	    if $logger;
	return undef;
    }
    my @rv = $rpc->data($hname);
    my $data = $rv[0];
    #use Data::Dumper; 
    #warn "RV: ", Dumper(@rv);

    # shift of the extra value used to seed the RPC 
    # headers.
    shift @$data; 
    return $data;
}

# This method only gets called on a _full_trash(). The server side will
# call a _full_trash() directly.
sub trash {
    my ($self, $msg, $logger, @locations) = @_;

    # It would be nice to be able to dispatch this request to all locations
    # in the same call but each location receives a different message ID (one
    # that is specific to that location).
    my $id = $msg->id;
    foreach my $loc (@locations) {
	my $host = PureMessage::ServerGroup::get_host($loc);
	$host->{port} = $host->{sync_port};
	$host->{proto} = $host->{sync_proto};
	my $loc_id = $msg->local_id($loc);

	my $rpc = PureMessage::RPC::ClientV2->new({$loc => $host});
	$rpc->call("QMan.trash", $loc_id);
	$rpc->wait();
	my $err = $rpc->error($host);
	$logger->log("ERROR", 
	    "Error dispatching the trashing of message $id to $host: $err")
	    if $err;
    }
}

sub forward {
    my ($self, $msg, $from, $to, %opts) = @_;

    $to = [$to] unless ref($to);
    
    my $logger = $opts{logger};
    my $loc = $msg->locations->[0];
    my $local_id = $msg->local_id($loc);
    my $host = PureMessage::ServerGroup::get_host($loc);
    $host->{port} = $host->{sync_port};
    $host->{proto} = $host->{sync_proto};
    my $rpc = PureMessage::RPC::ClientV2->new({$loc => $host});

    $rpc->call("QMan.forward", $local_id, $from, @$to);
    $rpc->wait();
    my $err = $rpc->error($host);
    if ($err) {
	$logger->log("ERROR",
	    "Error dispatching the forwarding of message $local_id to $loc: $err");
    }
    else {
	$logger->log("INFO",
	    "Dispatched the forwarding of message $local_id to $loc OK");
    }
}

1;

