package PureMessage::Queue;

use strict;
use base 'Exporter';
our @EXPORT_OK = qw(queue_mail);
use PureMessage::MessageInfo;
use HTTP::Date;
use Carp;

my $q;
sub _q {
    return $q ||= do {
	require PureMessage::Quarantine;
	PureMessage::Quarantine->new;
    }
}

sub queue_mail {
    my %opts = @_;
    
    my $tmp; 
    
    if ($opts{tempfile}) {
        $tmp = $opts{tempfile};
    }
    else {
        $tmp = _q()->new_tempfile;
        if (ref $opts{message}) {
            my $msg = $opts{message};
            my $buf;
            $msg->seek(0, 0) if $msg->can("seek");
            while ($msg->read($buf, 4096)) {
                $tmp->print($buf);
            }
        }
        elsif ($opts{message}) {
            $tmp->print($opts{message});
        }
    }

    my $from = $opts{from};
    my @to;
    if (ref($opts{to}) eq "ARRAY") {
	push(@to, grep(defined $_, @{$opts{to}}));
    }
    elsif (defined $opts{to}) {
	push(@to, $opts{to});
    }
    
    if ($opts{minfo}) {
        # make a local copy of the minfo - don't mess with the 
        # user's copy.
        my $minfo = PureMessage::MessageInfo->new($opts{minfo}->items);
        $minfo->value('Reason', 'Queue') unless lc($minfo->value('Reason')) eq 'route';
        $minfo->value('Date', HTTP::Date::time2str());
        $minfo->value('Envelope-To', @to) if @to;
        $minfo->value('Envelope-From', $from) if defined $from;
        $minfo->value('Store-Meta', 'No');
        $tmp->store($minfo);
        
    }
    else {
        croak 'Missing "to" parameter' unless @to;
        croak 'Missing "from" parameter' unless defined $from;

        $tmp->store(Reason => "Queue",
                    'Store-Meta' => 'No',
                    Date => HTTP::Date::time2str(),
                    'Envelope-From' => $from,
                    (map { ('Envelope-To' => $_) } @to),
                    );
    }
}

1;

__END__

=head1 NAME

PureMessage::Queue - The PureMessage mail queue

=head1 SYNPOSIS

  use PureMessage::Queue qw(queue_mail);

  queue_mail(
    from => 'sender@domain.com',
    to   => 'recipient@company.com',
    message => <<EOT);
  From: Sender <sender@domain.com>
  Subject: Hi there

  This is the message body.
  EOT

=head1 DESCRIPTION

The C<PureMessage::Queue> is the inteface to the PureMessage message
queue.  In its current form it only supports adding message to the
queue.

The following function is provided but not exported by default:

=over

=item queue_mail( %args )

The following arguments must be passed to queue_mail() as key/value pairs:

=over

=item from

The envelope sender address. See the ADDRESSES section in the C<Net::SMTP>
man page for information on supported address formats.  queue_mail() will croak
if this parameter is not defined and the minfo option was not specified.

=item to

The envelope recipient address or addresses. This can be either an array 
reference to a list of addresses or a single address. The address format
rules are the same as the from address.  queue_mail() will croak if this
parameter is not defined and the minfo option was not specified.

=item message

The body of the message. This can be a string containing the entire
message, an object reference where the object has an interface that
supports read() and optionally seek() with the same behaviour as
perl's built in file handles.  The C<PureMessage::MessageFile> class
support this interface.

It is the caller's responsibility to provide a valid e-mail message.

=item minfo

Pass a whole minfo block to be used when storing the message.  If this option is 
used, the 'to' and 'from' options are optional, but they will override the 
Envelope-To and Envelope-From values in the minfo if they are present.

=item tempfile

Instead of passing the text via the 'message' option, specify a MessageStore 
tempfile containing the message text.

=back

=back

=head1 SEE ALSO

L<pmx-queue>
