package PureMessage::QueueRunner;

use strict;
use warnings;
use Carp qw(croak);
use base qw(Exporter);

use Email::Received qw(parse_received);
use PureMessage::Config qw(%config expand_tmpl_filename);
use PureMessage::QueueRunner::Utils qw(xlog xdie);
use PureMessage::Quarantine;
use PureMessage::_sys;
use PureMessage::Queue qw(queue_mail);
use PureMessage::Template qw(expand_template);
use PureMessage::Queue qw(queue_mail);
use ActiveState::File::Atomic;
use DBI;

(my $progname = $0) =~ s,.*/,,;
my $hostname = PureMessage::_sys::gethostname() || "some-puremessage-host";

# Force a dir MessageStore - we shouldn't need a database to do a queue run.  We don't 
# supply a quarantine spec (let the quarantine figure it out) - hence the undef
my $q = PureMessage::Quarantine->new(undef, force_type => 'dir');

sub new {
    my ($class, $opts) = @_;

    $opts ||= {};
    
    my ($scheme) = split(/:/, $config{mail_sender}, 2);
    die "Unknown mail_sender scheme \"$scheme\"!\n" if $scheme ne 'smtp';

    # hostname?
    my $self = bless $opts, $class;
    return $self;
}

sub _show_message {
    my $m = shift;

    require HTTP::Date;
    print "#", $m->id;
    for ($m->minfo_value("Envelope-To")) {
	print " <$_>";
    }
    print " (", $m->message_size, " bytes)";
    my $date = HTTP::Date::str2time($m->minfo_value("Date"));
    print " ", HTTP::Date::time2iso($date);
    print "\n";

    require ActiveState::Duration;
    my $now = time;
    my $last_msg = "";
    for (reverse $m->minfo_value("Delivery-Failed")) {
	my($t, $msg) = split(/;\s*/, $_, 2);
	$t = HTTP::Date::str2time($t);
	if ($last_msg && $last_msg eq $msg) {
	    $msg = ' - - -';
	}
	else {
	    $last_msg = $msg;
	}
	print "\t", ActiveState::Duration::ago_eng($now - $t), ": $msg\n";
    }
    print "\n";
}

sub _qmail {
    # single message
    my $msg = shift;

    # from_addr - same for each message
    my $from = shift;

    my %opts = @_;

    # to_addr or array of to_addrs - destination(s) for each message
    my $to_addr = $opts{to};
    if (ref $to_addr and ref $to_addr ne 'ARRAY') {
        my $r = ref $to_addr;
        die "parameter 'to_addr' must be a scalar or an arrayref, not a '$r'";
    }

    # get the to address from the parameters or from the envelope
    # information
    my $to;
    if ($to_addr) {
        $to = (ref $to_addr) ? $to_addr : [$to_addr];
    }
    else {
        $to = [$msg->minfo_value("Envelope-To")];
    }

    # Add a copy of the local message into the mail queue
    my $l_msg = $msg;
    if ($msg->can("get_local_message")) {
        $l_msg = $msg->get_local_message;
    }
    log_message($msg, "Requeuing message for: @$to");
    queue_mail(from => $from, to => $to, message => $l_msg);

    return $to;
}

sub list {
    my ($self) = @_;
    
    my $n = 0;

    if (my $counter = eval { $q->counter("pmx-queue", timeout => 3) }) {
	my $prev_last_id = $counter->value || 0;
	
        # XXX: Why would this need undeffed immediately? it should be destroyed at the
        # end of this block
        #undef($counter);

	$q->message_scan(sub {
	    my $m = shift;
	    if (grep { $_ eq "Queue" || $_ eq "Route" } ($m->minfo_value("Reason")) ) {
		_show_message($m);
		$n++;
	    }
	}, "cur", $prev_last_id+1);
    }
    else {
	print "Some queued messages may not be shown since 'pmx-queue run' is active.\n\n";
    }

    if ( grep { $_ eq "Queue" } $q->folders() ) {
	$n += $q->message_scan(\&_show_message, "Queue");
    }

    my $s = ($n == 1) ? "" : "s";
    print "($n message$s)\n";
    return $n;
}

sub process_queued_messages {
    my ($self, $opts) = @_;
    my $count = 0;

    $opts->{quarantine_secret} = $config{queuerunner}{quarantine_secret};

    $count += $self->retry_queued_messages($opts);
    $count += $self->send_queued_messages($opts);
    return $count;
}

sub retry_message {
    my ($self, $m, $message_count, $opts) = @_;
    $m = $q->message($m) unless (ref($m));

    my $min_backoff_time   = $opts->{'min-backoff-time'} 
                            || $config{queuerunner}{'min-backoff-time'};
    my $max_backoff_time   = $opts->{'max-backoff-time'} 
                            || $config{queuerunner}{'max-backoff-time'};
    my $max_queue_lifetime = $opts->{'max-queue-lifetime'}
                            || $config{queuerunner}{'max-queue-lifetime'};

    $min_backoff_time   = _dur_parse($min_backoff_time);
    $max_backoff_time   = _dur_parse($max_backoff_time);
    $max_queue_lifetime = _dur_parse($max_queue_lifetime);

    # should we filter it out?    
    my @fail = $m->minfo_value("Delivery-Failed");
    if (@fail) {	
        require HTTP::Date;

        return if defined $opts->{"min-failures"}
            && $opts->{"min-failures"} > @fail;
	    return if defined $opts->{"max-failures"}
            && $opts->{"max-failures"} < @fail;

        my $last_attempt = $fail[-1] or return;
        my $first_attempt = $fail[0];
        $last_attempt =~ s/;.*//;
        $first_attempt =~ s/;.*//;

        my $life_time = time - HTTP::Date::str2time($first_attempt);
        if ($life_time > $max_queue_lifetime) {
            _bounce_on_failure($m, $fail[-1]);
            $m->file_in("Failed");
            return;
        }

        # double min backoff time untill reach max backoff time
        my $interval = time - HTTP::Date::str2time($last_attempt);
        my $need_to_wait = (1 << $#fail) * $min_backoff_time;
        $need_to_wait = $max_backoff_time if $need_to_wait > $max_backoff_time;
        return if $interval < $need_to_wait;
    }

    # try it then
    $$message_count++;
    $self->mail_message($m, $opts);
}

sub create_workers {
    my ($self, $minions) = @_;
    die "create_workers(): need number of workers"
        unless $minions and $minions > 0;
    my $me = 'PureMessage::QueueRunner';
    open $self->{workers}[$_], '|-', qq{perl -M$me -e'$me->new->do_work'}
        for 0..$minions-1;

    for my $worker (@{$self->{workers}}) {
        my $current_fh = select($worker);
        $| = 1;
        select($current_fh);
    }

    $self->{next_worker} = 0;    
}

sub do_work {
    my ($self) = @_;

    my $quarantine_secret = $config{queuerunner}{quarantine_secret};

    eval { $q->message_writable(1) };
    xdie("message_writable error: $@") if $@;

    while (my $id = <STDIN>) {
        chomp($id);
        my $m;
        eval {$m = $q->message($id) or next };
        xdie("message error (id='$id'): $@") if $@;
        eval { $self->mail_message($m, { quarantine_secret => $quarantine_secret }) };
        xlog("ERR", "mail_message failed on $id: $@") if $@;
    }
}

sub send_queued_messages {
    my ($self, $opts) = @_;
    $opts ||= {};
    return 0 if defined $opts->{"min-failures"};

    my $progname = $self->{progname};
    $q->message_writable(1);

    my $counter = eval {
        $q->counter('pmx-queue', writable => 1, create => 1, timeout => 5);
    } or xdie "$progname: Can't obtain queue lock.\n" .
             "Another instance of 'pmx-queue run' is probably running.\n";
    my $prev_last_id = $counter->value || 0;

    xlog("DEBUG", "Scanning cur from $prev_last_id");
    my $last_id;
    my $message_count = 0;
    $q->message_scan(sub {
        my $m = shift;
        $last_id = $m->id;
        if (grep { $_ eq "Queue" || $_ eq "Route" } $m->minfo_value("Reason")) {
            $message_count++;
            if (defined $self->{workers}) {
                # If workers is defined (create_workers was called),
                # use the worker processes to send the messages.
                $self->{next_worker} =
                    ($self->{next_worker} + 1) % @{$self->{workers}};
                my $fh = $self->{workers}[$self->{next_worker}];
                my $id = $m->id;
                print $fh "$id\n";
            }
            else {
                # Send messages in this process.
                $self->mail_message($m, $opts);
            }
        }
    }, "cur", $prev_last_id+1);

    if (defined($last_id)) {
        xlog("DEBUG", "Save last_queue_id as $last_id");
        $counter->value($last_id);
        $counter->save;
    }
    xlog("DEBUG", "Done scanning cur from $prev_last_id, found $message_count messages");
    return $message_count;
}

sub _dur_parse {
    my $range = shift;
    $range =~ /^(\d+)(s|m|h|d|w)?$/;
    my $t_period = $2;
    $t_period = 's' unless $2; # by default the unit is 'seconds'
    # convert all times into seconds for internal use
    $range = ($t_period eq 's') ? 1 :
                  ($t_period eq 'm') ? 60 :
                        ($t_period eq 'h') ? 3600 :
                                ($t_period eq 'd') ? 86400 : 604800;
    $range *= $1;
    return $range;
}

sub _bounce_on_failure {
    my ($m, $f) = @_;

    if ($m->minfo_value("Bounce-On-Failure")
     && $m->minfo_value("Bounce-On-Failure") eq "Yes") {
        my $tmpl_file = expand_tmpl_filename("bounce-on-failure.tmpl");    
        my $mail_text = expand_template(
            ActiveState::File::Atomic->new($tmpl_file)->slurp,
            vars => {
                'failure'       => $f,
                'msg_from'      => $m->minfo_value("Envelope-From"),
                'msg_to'        => join(',', $m->minfo_value('Envelope-To')),
                'msg_subject'   => $m->header("Subject") || '',
                'msg_sent_date' => $m->header("Date")    || '',
            }
        );        
        queue_mail(from => 'postmaster',
                   to => $m->minfo_value("Bounce-To") ||
                         $m->minfo_value("Envelope-From"),
                   message => $mail_text);
    }
}

sub retry_queued_messages {
    my ($self, $opts) = @_;
    $opts ||= {};
    my $progname = $self->{progname};
    $q->message_writable(1);

    xlog("DEBUG", "Scanning entire Queue folder");
    my $message_count = 0;
    eval {
	$q->message_scan(sub {
            my $m = shift;
            $self->retry_message($m, \$message_count, $opts);
	}, "Queue");
    };
    warn $@ if $@ && $@ !~ /^No such folder 'Queue'/;
    xlog("DEBUG", "Done scanning entire Queue folder, found $message_count messages");
    return $message_count;
}

sub mail_message {
    my ($self, $m, $opts) = @_;
    
    # Only SMTP supported
    my $smtp;
    eval {
        $smtp = new_smtp($config{mail_sender});
    };
    if ($@) {
        my $err = $@;
        $err =~ s/\n/ /g;
        $err =~ s/\s+ / /g;
        $err =~ s/\s+$//;
        my $minfo = $m->minfo;
        $minfo->append_items("Delivery-Failed" => 
                             HTTP::Date::time2str() . "; 101: $err");
        $m->minfo($minfo);
        $m->file_in("Queue") unless $m->folder eq "Queue";
        log_message($m, $err);
        return;
    }
   
    # Message logging variables
    my $mail_rcpt_failures = { };
    my @mail_rcpt_success = ( );
    my $mail_accept_msg = '';
    my $status = '';

    eval {
        mail_verify_sender($smtp, $m);
        @mail_rcpt_success = mail_verify_recipients($smtp, $m, $mail_rcpt_failures);
        if(scalar @mail_rcpt_success) {
            $mail_accept_msg  = mail_send_data($smtp, $m, $opts);
        }
    };

    # Permanent SMTP send failure
    if ($@) {
        my $err = $@;
        $err =~ s/\n/ /g;
        $err =~ s/\s+ / /g;
        $err =~ s/\s+$//;
        my $code;
        ($code, $err) = split (':', $err, 2);
        # print "Message $id: $code: $err\n";
        require HTTP::Date;
        my $minfo = $m->minfo;
        $minfo->append_items("Delivery-Failed" =>
                             HTTP::Date::time2str() . "; $code: $err");
        $m->minfo($minfo);
        if ($code =~ /^5/) {
         $status = "rejected ($code $err)";
            _bounce_on_failure($m, $status);
            $m->file_in("Failed");
        }
        else {
         $status = "deferred ($code $err)";
            $m->file_in("Queue") unless $m->folder eq "Queue";
        }
    } 
    elsif(mail_rcpt_all_failures($mail_rcpt_failures)) { 
        # Here we deal with messages that had some sort of recipient failure

        # If there are multiple recipient status, 
        #   i.e. the message has been sent for some recipients and failed for 
        #        others
        # then we need to requeue/split the message for the failed recipients
        # If all the recipients have the same status, we can simply 'File' 
        # the message as a whole
        my @rcpts;
        my $from = $m->minfo_value("Envelope-From");

        # Temporary recipient failures
        if(mail_rcpt_tmp_failures($mail_rcpt_failures)) {        
            $status = "deferred: temporary recipient failures (" 
                    . mail_rcpt_tmp_failures($mail_rcpt_failures) . ")";
           
            # Rewrite Envelope-To to only contain temporary failed recipients
            for (keys %$mail_rcpt_failures) {
                push @rcpts, $_ if $mail_rcpt_failures->{$_} =~ /^4/;
            }
            $m->minfo_value("Envelope-To", @rcpts);

            if (scalar @mail_rcpt_success || mail_rcpt_permanent_failures($mail_rcpt_failures)) {
                _qmail($m, $from, to => \@rcpts);
            }
            else {
                $m->file_in("Queue") unless $m->folder eq "Queue";	    
            }
            log_message($m, $status);
        }
        # Permanent recipient failures
        if(mail_rcpt_permanent_failures($mail_rcpt_failures)) {        
            $status = "rejected: permament recipient failures (" 
                    . mail_rcpt_permanent_failures($mail_rcpt_failures) . ")";

            # Rewrite Envelope-To to only contain permanent failed recipients
            @rcpts = ( );
            for (keys %$mail_rcpt_failures) {
                push @rcpts, $_ if $mail_rcpt_failures->{$_} =~ /^5/;
            }
            $m->minfo_value("Envelope-To", @rcpts);

            if (scalar @mail_rcpt_success || mail_rcpt_tmp_failures($mail_rcpt_failures)) {
                _qmail($m, $from, to => \@rcpts);
            }
            else {
                $m->file_in("Failed");
            }
            log_message($m, $status);
        }

        # If there were valid recipients the message also needs to be filed in
        # the 'Sent' folder
        if(scalar @mail_rcpt_success) {
            $m->minfo_value("Envelope-To", @mail_rcpt_success);
            # Message sent successfully
            $status = "sent (250 " . $mail_accept_msg . ")";
            $m->file_in("Sent");
            log_message($m, $status);
        }
    }
    else {
        # Message sent successfully
        $m->file_in("Sent");
        $status = "sent (250 " . $mail_accept_msg . ")";
        log_message($m, $status);
    }
}

sub log_message {
    my ($m, $status) = @_;

    my @loginfo;
    # Parse the postfix id
    my $pfid;
    if (my $recv_hdr = $m->header('Received', 0)) {	
        my $parsed_hdr = parse_received($recv_hdr);		
	    $pfid = $parsed_hdr->{id};
    }
    
    # Log if we are a route-copy, we need to log extra 
    # information for logsearch to read and display.
    my $route_copy = $m->minfo_value('Route-Copy');
    if (defined $route_copy and $route_copy eq 'Yes') {        
        push @loginfo, "${pfid}_COPY:" if defined $pfid;

    } else {
        # Normal route message uses the original pfid
        push @loginfo, "$pfid:" if defined $pfid;
    }
    
    # Quarantine ID (same as in message_log)
    if (my $qid = $m->minfo_value('Queue-Id')) {
        push @loginfo, "q=$qid";
    }

    # Envelope from
    if (my $from = $m->minfo_value('Envelope-From')) {
        push @loginfo, "from=$from";
    }

    # Envelope to(s)
    if (my $to = join ',', $m->minfo_value('Envelope-To')) {
        push @loginfo, "to=$to";
    }

    # Client relay
    if (my $client = $m->minfo_value('Relay')) {
        $client =~ s/.*\[(.+)\]/$1/g;
        push @loginfo, "client=$client";
    }

    # Subject - only log this if we're routing the message
    if (defined $route_copy and $route_copy eq 'Yes') {        
        if (my $subject = $m->header('Subject') || '') {
            $subject =~ s/_/__/g;
            $subject =~ s/\s/_/g;
            push @loginfo, "subject=$subject";
        }
    }

    # Message ID
    if (my $msgid = $m->header('Message-Id')) {
          $msgid =~ s/<(.+)>/$1/sg;
          push @loginfo, "msgid=$msgid";
    }

    # Global ID
    if (my $gid = $m->minfo_value('Global-ID')) {
        push @loginfo, "gid=$gid";
    }

    # Log the downstream (route) relay if it exists
    if (defined(my $relay = $m->minfo_value("Destination-Server"))) {
        push @loginfo, "relay=$relay";
    }
        
    xlog("INFO", "@{loginfo} status=$status");
}

sub mail_rcpt_fail {
    my ($to, $obj, $m, $mail_rcpt_failures) = @_;
    $mail_rcpt_failures->{$to} = $obj->code.":".$obj->message;
    mail_smtp_warn($obj, $m);
}

sub mail_rcpt_tmp_failures {
    my $mail_rcpt_failures = shift;
    # Return all recipients from this message that had temp failures
    my $tmp = "";
    for (keys %$mail_rcpt_failures) {
        $tmp .= "$_=" . $mail_rcpt_failures->{$_} . ";" if $mail_rcpt_failures->{$_} =~ /^4/;
    }
    return $tmp;
}

sub mail_rcpt_permanent_failures {
    my $mail_rcpt_failures = shift;
    # Return all recipients from this message that had permanent failures
    my $tmp = "";
    for (keys %$mail_rcpt_failures) {
        $tmp .= "$_=" . $mail_rcpt_failures->{$_} . ";" if $mail_rcpt_failures->{$_} =~ /^5/;
    }
    return $tmp;
}

sub mail_rcpt_all_failures {
    my $mail_rcpt_failures = shift;
    # Return all recipients from this message that failures
    my $tmp = "";
    for (keys %$mail_rcpt_failures) {
        $tmp .= "$_=" . $mail_rcpt_failures->{$_} . ";";
    }
    return $tmp;
}

sub log_smtp_msg {
    # This private sub will be called by either mail_smtp_die or mail_smtp_warn.
    # The type is either 'die' or 'warn'.
    my ($obj, $type) = @_;

    # Net::SMTP will return a status of either "599" or "000" when a connection
    # closes in mid-message. We want this to be treated as a temporary failure
    # so the message is requeued instead of destroyed.
    my $code = $obj->code || 555;
    my $msg;
    if($code eq '000') {
        $code = '499';
        $msg = $obj->message || "$!" || '(Unknown connection failure)';
        xlog("DEBUG", "mail_smtp_${type}: code='$code' msg='$msg'");
    } elsif($code eq '599') {
        $code = '499';
        $msg = $obj->message || '(Unknown connection failure)';
        xlog("DEBUG", "mail_smtp_${type}: code='$code' msg='$msg'");
    } else {
        $msg = $obj->message || '(Unknown error)';
        xlog("DEBUG", "mail_smtp_${type}: code='$code' msg='$msg'");
    }

    return ($msg, $code);
}

sub mail_smtp_check {
    my ($obj, $m) = @_;

    # Only die on 4 or 5 level smtp error codes
    if($obj->code =~ /^[45]/) {
        mail_smtp_die($obj,$m);
    }
    else {
        mail_smtp_warn($obj,$m);
    }
}

sub mail_smtp_die {
    my ($obj, $m) = @_;

    my (undef, $file, $line) = caller;
    my ($msg, $code) = log_smtp_msg($obj, 'die');
    my $die_msg = "$code:$msg at $file line $line\n";
    xlog("WARNING", "mail_smtp_die: $die_msg");
    die $die_msg;
}

sub mail_smtp_warn {
    my ($obj, $m) = @_;

    my (undef, $file, $line) = caller;
    my ($msg, $code) = log_smtp_msg($obj, 'warn');
    my $warn_msg = "$code:$msg at $file line $line\n";
    xlog("WARNING", "mail_smtp_warn: $warn_msg");
}

sub mail_received_line {
    my $m = shift;
    # construct Received field for PM quarantine/queue
    my $sm_host = $m->minfo_value('Sendmail-Host') || $hostname;
    my $id      = $m->minfo_value('Global-Id') || $m->id;
    my @env_to  = $m->minfo_value('Envelope-To');
    my $date    = $m->minfo_value('Date');
    my $fqdn    = $m->minfo_value('Milter-Host')
		    || $m->minfo_value('Sendmail-Host')
		        || $hostname;
    my $recvd = "Received: from $sm_host\r\n\tby $fqdn with queue id $id";
    $recvd   .= @env_to > 1 ? ";\r\n\t" : "\r\n\tfor $env_to[0]; ";
    $recvd   .= "$date\r\n";
    $recvd;
}

sub mail_edited_headers {
    my ($m, $opts) = @_;
    my $recvd = mail_received_line($m);
    # insert our Received header.
    # put our header in first, unless the return-path header is first,
    # in which case put it after that.
    my $hdr = $m->header;
    my @hdrs;
    push @hdrs, $_ foreach $hdr =~ /([^\n]*(?:\n\s+[^\n]*)*\n)/g;
    my $insert_before = 0;
    $insert_before = 1 if $hdrs[0] =~ /^return-path:/i;
    @hdrs = (@hdrs[0..$insert_before-1], $recvd, @hdrs[$insert_before..$#hdrs]);

    # send quarantine secret if necessary
    if (!$opts->{'no-secret'} && $opts->{'quarantine_secret'}) {

        if (my $secret = $config{quarantine_secret}) {
            unshift(@hdrs, "X-PMX-Quarantine-Approved: $secret\r\n");
        }
    }

    # seek to the end of the header block
    $m->seek(length($hdr));

    return(@hdrs);
}

sub new_smtp {
    my $sender = shift;
    require Net::SMTP;
    my ($scheme, $host, $port) = split(':', $sender);
    $port ||= 25;
    my $smtp = Net::SMTP->new($host, Port => $port, Timeout => 10);
    die "connect to $host: Operation timed out\n"
        unless $smtp;
    return $smtp;
}

sub mail_verify_sender {
    my ($smtp, $m) = @_;
 
    my $from = $m->minfo_value("Envelope-From");
    $from = "<>" if $from eq "";
    xlog("DEBUG", "mail_send_smtp: from='$from'");

    $smtp->mail($from) || mail_smtp_die($smtp, $m);
}

sub mail_verify_recipients {
    my ($smtp, $m, $mail_rcpt_failures) = @_;

    my %valid_recipients = map { $_ => 1 } $m->minfo_value("Envelope-To");
    for ($m->minfo_value("Envelope-To")) {
        xlog("DEBUG", "mail_send_smtp: to='$_'");
    	$smtp->to($_) || mail_rcpt_fail($_, $smtp, $m, $mail_rcpt_failures);
    }

    # Remove invalid recipients
    delete $valid_recipients{$_} for keys %$mail_rcpt_failures;

    return keys %valid_recipients;
}

sub mail_send_data {
    my ($smtp, $m, $opts) = @_;
   
    xlog("DEBUG", "mail_send_smtp: data");
    $smtp->data()|| mail_smtp_check($smtp, $m);

    my(@hdrs) = (mail_edited_headers($m, $opts));

    # send the header block
    $smtp->datasend(@hdrs);

    # send the rest of the message
    my $buf;
    my $maybe_cr = "";
    my $block_size = $config{block_size} || 1024;
    while ($m->read($buf, $block_size)) {
        # avoid breaking up \r\n
        # if the last character of the block is a \r,
        # remove it and append it to the beginning of the
        # next block.
        $buf = $maybe_cr . $buf if $maybe_cr;
        $maybe_cr = ($buf =~ s/\r\z//) ? "\r" : "";
        $smtp->datasend($buf) || mail_smtp_check($smtp, $m);
    }
    $smtp->datasend($maybe_cr) if $maybe_cr;

    xlog("DEBUG", "mail_send_smtp: data end");
    $smtp->dataend || mail_smtp_check($smtp, $m);

    # Store Accept Message
    my $accept_msg = $smtp->message;
    $smtp->quit;

    $accept_msg =~ s/^Enter mail[,.] [Ee]nd with (?:"\."|the \. character) on a line by itself\.?\s*//;
    $accept_msg =~ s{^Start mail input; end with <CRLF>\.<CRLF>\s*}{};
    $accept_msg =~ s{End data with <CR><LF>\.<CR><LF>\s*}{};
    $accept_msg =~ s{\s+$}{};

    return $accept_msg;
}
1;
